"""This file is part of the binoculars project.

  The BINoculars library is free software: you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3 of
  the License, or (at your option) any later version.

  The BINoculars library is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the hkl library.  If not, see
  <http://www.gnu.org/licenses/>.

  Copyright (C) 2015-2021 Synchrotron SOLEIL
                          L'Orme des Merisiers Saint-Aubin
                          BP 48 91192 GIF-sur-YVETTE CEDEX

  Copyright (C) 2012-2015 European Synchrotron Radiation Facility
                          Grenoble, France

  Authors: Willem Onderwaater <onderwaa@esrf.fr>
           Picca Frédéric-Emmanuel <picca@synchrotron-soleil.fr>

"""
from typing import Any, Dict, NamedTuple, Optional, Tuple

import numpy
import math
import os
import sys

import pyFAI

from enum import Enum
from math import cos, sin

from gi.repository import GLib
import gi
gi.require_version("Hkl", "5.0")
from gi.repository import Hkl

from h5py import Dataset, File
from numpy import ndarray
from numpy.linalg import inv
from pyFAI.detectors import ALL_DETECTORS

from .soleil import (
    DatasetPathContains,
    DatasetPathOr,
    DatasetPathWithAttribute,
    HItem,
    get_dataset,
    get_nxclass,
    node_as_string,
)
from .. import backend, errors, util
from ..util import ConfigSection

# TODO
# - Angles delta gamma. nom de 2 ou 3 moteurs. omega puis delta
#   gamma pour chaque pixels.

# - aller cherche dans le fichier NeXuS le x0, y0 ainsi que le sdd.

# - travailler en qx qy qz, il faut rajouter un paramètre optionnel
# - qui permet de choisir une rotation azimuthal de Qx Qy.

###################
# Common methodes #
###################

WRONG_ATTENUATION = -100


class Diffractometer(NamedTuple):
    name: str  # name of the hkl diffractometer
    ub: ndarray  # the UB matrix
    geometry: Hkl.Geometry  # the HklGeometry


def get_diffractometer(hfile: File):
    """ Construct a Diffractometer from a NeXus file """
    node = get_nxclass(hfile, "NXdiffractometer")

    name = node_as_string(node["type"][()])
    if name.endswith("\n"):
        # remove the last "\n" char
        name = name[:-1]

    try:
        ub = node["UB"][:]
    except AttributeError:
        ub = None

    factory = Hkl.factories()[name]
    geometry = factory.create_new_geometry()

    # wavelength = get_nxclass(hfile, 'NXmonochromator').wavelength[0]
    # geometry.wavelength_set(wavelength)

    return Diffractometer(name, ub, geometry)


class Sample(NamedTuple):
    a: float
    b: float
    c: float
    alpha: float
    beta: float
    gamma: float
    ux: float
    uy: float
    uz: float
    ub: ndarray
    sample: Hkl.Sample


def get_sample(hfile, config):
    """ Construct a Diffractometer from a NeXus file """
    node = get_nxclass(hfile, "NXdiffractometer")

    def get_value(node, name, default, overwrite):
        if overwrite is not None:
            v = overwrite
        else:
            v = default
            try:
                v = node[name][()][0]
            except AttributeError:
                pass
        return v

    # hkl default sample
    a = get_value(node, "A", 1.54, config.a)
    b = get_value(node, "B", 1.54, config.b)
    c = get_value(node, "C", 1.54, config.c)
    alpha = get_value(node, "alpha", 90, config.alpha)
    beta = get_value(node, "beta", 90, config.beta)
    gamma = get_value(node, "gamma", 90, config.gamma)
    ux = get_value(node, "Ux", 0, config.ux)
    uy = get_value(node, "Uy", 0, config.uy)
    uz = get_value(node, "Uz", 0, config.uz)

    sample = Hkl.Sample.new("test")
    lattice = Hkl.Lattice.new(
        a, b, c, math.radians(alpha), math.radians(beta), math.radians(gamma)
    )
    sample.lattice_set(lattice)

    parameter = sample.ux_get()
    parameter.value_set(ux, Hkl.UnitEnum.USER)
    sample.ux_set(parameter)

    parameter = sample.uy_get()
    parameter.value_set(uy, Hkl.UnitEnum.USER)
    sample.uy_set(parameter)

    parameter = sample.uz_get()
    parameter.value_set(uz, Hkl.UnitEnum.USER)
    sample.uz_set(parameter)

    ub = hkl_matrix_to_numpy(sample.UB_get())

    return Sample(a, b, c, alpha, beta, gamma, ux, uy, uz, ub, sample)


class Detector(NamedTuple):
    name: str
    detector: Hkl.Detector


def get_detector(hfile, h5_nodes):
    detector = Hkl.Detector.factory_new(Hkl.DetectorType(0))
    images = h5_nodes["image"]
    s = images.shape[-2:]
    if s == (960, 560) or s == (560, 960):
        det = Detector("xpad_flat", detector)
    elif s == (1065, 1030):
        det = Detector("eiger1m", detector)
    elif s == (120, 560):
        det = Detector("imxpads70", detector)
    elif s == (256, 257):
        det = Detector("ufxc", detector)
    else:
        det = Detector("imxpads140", detector)

    return det


class Source(NamedTuple):
    wavelength: float


def get_source(hfile):
    wavelength = None
    node = get_nxclass(hfile, "NXmonochromator")
    for attr in ["wavelength", "lambda"]:
        try:
            wavelength = node[attr][0]
        except KeyError:
            pass
        except IndexError:
            pass

    return Source(wavelength)


class DataFrame(NamedTuple):
    diffractometer: Diffractometer
    sample: Sample
    detector: Detector
    source: Source
    h5_nodes: Dict[str, Dataset]


def dataframes(hfile, data_path, config):
    h5_nodes = {k: get_dataset(hfile, v) for k, v in data_path.items()}
    diffractometer = get_diffractometer(hfile)
    sample = get_sample(hfile, config)
    detector = get_detector(hfile, h5_nodes)
    source = get_source(hfile)

    yield DataFrame(diffractometer, sample, detector, source, h5_nodes)


def get_ki(wavelength):
    """
    for now the direction is always along x
    """
    TAU = 2 * math.pi
    return numpy.array([TAU / wavelength, 0, 0])


def normalized(a, axis=-1, order=2):
    l2 = numpy.atleast_1d(numpy.linalg.norm(a, order, axis))
    l2[l2 == 0] = 1
    return a / numpy.expand_dims(l2, axis)


def hkl_matrix_to_numpy(m):
    M = numpy.empty((3, 3))
    for i in range(3):
        for j in range(3):
            M[i, j] = m.get(i, j)
    return M


def M(theta, u):
    """
    :param theta: the axis value in radian
    :type theta: float
    :param u: the axis vector [x, y, z]
    :type u: [float, float, float]
    :return: the rotation matrix
    :rtype: numpy.ndarray (3, 3)
    """
    c = cos(theta)
    one_minus_c = 1 - c
    s = sin(theta)
    return numpy.array(
        [
            [
                c + u[0] ** 2 * one_minus_c,
                u[0] * u[1] * one_minus_c - u[2] * s,
                u[0] * u[2] * one_minus_c + u[1] * s,
            ],
            [
                u[0] * u[1] * one_minus_c + u[2] * s,
                c + u[1] ** 2 * one_minus_c,
                u[1] * u[2] * one_minus_c - u[0] * s,
            ],
            [
                u[0] * u[2] * one_minus_c - u[1] * s,
                u[1] * u[2] * one_minus_c + u[0] * s,
                c + u[2] ** 2 * one_minus_c,
            ],
        ]
    )

###############
# Projections #
###############


class SurfaceOrientation(Enum):
    VERTICAL = 1
    HORIZONTAL = 2


class PDataFrame(NamedTuple):
    pixels: ndarray
    k: float
    ub: Optional[ndarray]
    R: ndarray
    P: ndarray
    index: int
    timestamp: int
    surface_orientation: SurfaceOrientation
    dataframe: DataFrame
    input_config: ConfigSection


class RealSpace(backend.ProjectionBase):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        pixels = pdataframe.pixels
        P = pdataframe.P
        timestamp = pdataframe.timestamp

        pixels_ = numpy.tensordot(P, pixels, axes=1)
        x = pixels_[1]
        y = pixels_[2]
        z = numpy.ones_like(x) * timestamp

        return (x, y, z)

    def get_axis_labels(self):
        return ("x", "y", "z")


class Pixels(backend.ProjectionBase):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        pixels = pdataframe.pixels

        return numpy.meshgrid(
            numpy.arange(pixels[0].shape[1]), numpy.arange(pixels[0].shape[0])
        )

    def get_axis_labels(self) -> Tuple[str]:
        return "x", "y"


class HKLProjection(backend.ProjectionBase):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        pixels = pdataframe.pixels
        k = pdataframe.k
        UB = pdataframe.ub
        R = pdataframe.R
        P = pdataframe.P

        if UB is None:
            raise Exception(
                "In order to compute the HKL projection, you need a valid ub matrix"
            )

        ki = [1, 0, 0]
        RUB_1 = inv(numpy.dot(R, UB))
        RUB_1P = numpy.dot(RUB_1, P)
        kf = normalized(pixels, axis=0)
        hkl_f = numpy.tensordot(RUB_1P, kf, axes=1)
        hkl_i = numpy.dot(RUB_1, ki)
        hkl = hkl_f - hkl_i[:, numpy.newaxis, numpy.newaxis]

        h, k, l = hkl * k

        return h, k, l

    def get_axis_labels(self) -> Tuple[str]:
        return "H", "K", "L"


class HKProjection(HKLProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        h, k, l = super(HKProjection, self).project(index, pdataframe)
        return h, k

    def get_axis_labels(self) -> Tuple[str]:
        return "H", "K"


class QxQyQzProjection(backend.ProjectionBase):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        pixels = pdataframe.pixels
        k = pdataframe.k
        R = pdataframe.R
        P = pdataframe.P
        surface_orientation = pdataframe.surface_orientation

        # TODO factorize with HklProjection. Here a trick in order to
        # compute Qx Qy Qz in the omega basis.
        if surface_orientation is SurfaceOrientation.VERTICAL:
            UB = numpy.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])

            if self.config.omega_offset is not None:
                UB = numpy.dot(UB, M(self.config.omega_offset, [0, 0, -1]))
        else:
            UB = numpy.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

            if self.config.mu_offset is not None:
                UB = numpy.dot(UB, M(self.config.mu_offset, [0, 0, -1]))

        # the ki vector should be in the NexusFile or easily extracted
        # from the hkl library.
        ki = [1, 0, 0]
        RUB_1 = inv(numpy.dot(R, UB))
        RUB_1P = numpy.dot(RUB_1, P)
        kf = normalized(pixels, axis=0)
        hkl_f = numpy.tensordot(RUB_1P, kf, axes=1)
        hkl_i = numpy.dot(RUB_1, ki)
        hkl = hkl_f - hkl_i[:, numpy.newaxis, numpy.newaxis]

        qx, qy, qz = hkl * k
        return qx, qy, qz

    def get_axis_labels(self) -> Tuple[str]:
        return "Qx", "Qy", "Qz"

    def parse_config(self, config) -> None:
        super(QxQyQzProjection, self).parse_config(config)

        # omega offset for the sample in degree then convert into radian
        omega_offset = config.pop("omega_offset", None)
        if omega_offset is not None:
            self.config.omega_offset = math.radians(float(omega_offset))
        else:
            self.config.omega_offset = None

        # omega offset for the sample in degree then convert into radian
        mu_offset = config.pop("mu_offset", None)
        if mu_offset is not None:
            self.config.mu_offset = math.radians(float(mu_offset))
        else:
            self.config.mu_offset = None


class QxQyIndexProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        timestamp = pdataframe.timestamp

        qx, qy, qz = super(QxQyIndexProjection, self).project(index, pdataframe)
        return qx, qy, numpy.ones_like(qx) * timestamp

    def get_axis_labels(self) -> Tuple[str]:
        return "Qx", "Qy", "t"


class QxQzIndexProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        timestamp = pdataframe.timestamp

        qx, qy, qz = super(QxQzIndexProjection, self).project(index, pdataframe)
        return qx, qz, numpy.ones_like(qx) * timestamp

    def get_axis_labels(self) -> Tuple[str]:
        return "Qx", "Qz", "t"


class QyQzIndexProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        timestamp = pdataframe.timestamp

        qx, qy, qz = super(QyQzIndexProjection, self).project(index, pdataframe)
        return qy, qz, numpy.ones_like(qy) * timestamp

    def get_axis_labels(self) -> Tuple[str]:
        return "Qy", "Qz", "t"


class QparQperProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        qx, qy, qz = super(QparQperProjection, self).project(index, pdataframe)
        return numpy.sqrt(qx * qx + qy * qy), qz

    def get_axis_labels(self) -> Tuple[str]:
        return "Qpar", "Qper"


class QparQperIndexProjection(QparQperProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        timestamp = pdataframe.timestamp

        qpar, qper = super(QparQperIndexProjection, self).project(index, pdataframe)
        return qpar, qper, numpy.ones_like(qpar) * timestamp

    def get_axis_labels(self) -> Tuple[str]:
        return "Qpar", "Qper", "t"


class Stereo(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        qx, qy, qz = super(Stereo, self).project(index, pdataframe)
        q = numpy.sqrt(qx * qx + qy * qy + qz * qz)
        ratio = qz + q
        xp = qx / ratio
        yp = qy / ratio
        return q, xp, yp

    def get_axis_labels(self) -> Tuple[str]:
        return "Q", "xp", "yp"


class QzPolarProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        qx, qy, qz = super(QzPolarProjection, self).project(index, pdataframe)
        phi = numpy.rad2deg(numpy.arctan2(qx, qy))
        q = numpy.sqrt(qx * qx + qy * qy + qz * qz)
        return phi, q, qz

    def get_axis_labels(self) -> Tuple[str]:
        return "Phi", "Q", "Qz"


class QyPolarProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        qx, qy, qz = super(QyPolarProjection, self).project(index, pdataframe)
        phi = numpy.rad2deg(numpy.arctan2(qz, qx))
        q = numpy.sqrt(qx * qx + qy * qy + qz * qz)
        return phi, q, qy

    def get_axis_labels(self) -> Tuple[str]:
        return "Phi", "Q", "Qy"


class QxPolarProjection(QxQyQzProjection):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        qx, qy, qz = super(QxPolarProjection, self).project(index, pdataframe)
        phi = numpy.rad2deg(numpy.arctan2(qz, -qy))
        q = numpy.sqrt(qx * qx + qy * qy + qz * qz)
        return phi, q, qx

    def get_axis_labels(self) -> Tuple[str]:
        return "Phi", "Q", "Qx"


class QIndex(Stereo):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        timestamp = pdataframe.timestamp

        q, qx, qy = super(QIndex, self).project(index, pdataframe)
        return q, numpy.ones_like(q) * timestamp

    def get_axis_labels(self) -> Tuple[str]:
        return "Q", "Index"

class AnglesProjection(backend.ProjectionBase):
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        # put the detector at the right position

        pixels = pdataframe.pixels
        geometry = pdataframe.dataframe.diffractometer.geometry
        detrot = pdataframe.input_config.detrot
        sdd = pdataframe.input_config.sdd

        try:
            axis = geometry.axis_get("eta_a")
            eta_a = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            eta_a = 0
        try:
            axis = geometry.axis_get("omega")
            omega0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            omega0 = 0
        try:
            axis = geometry.axis_get("delta")
            delta0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            delta0 = 0
        try:
            axis = geometry.axis_get("gamma")
            gamma0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            gamma0 = 0

        P = M(math.radians(eta_a), [1, 0, 0])
        if detrot is not None:
            P = numpy.dot(P, M(math.radians(detrot), [1, 0, 0]))

        x, y, z = numpy.tensordot(P, pixels, axes=1)

        delta = numpy.rad2deg(numpy.arctan(z / sdd)) + delta0
        gamma = numpy.rad2deg(numpy.arctan(y / sdd)) + gamma0
        omega = numpy.ones_like(delta) * omega0

        return (delta, gamma, omega)

        # # on calcule le vecteur de l'axes de rotation de l'angle qui
        # # nous interesse. (ici delta et gamma). example delta (0, 1,
        # # 0) (dans le repere du detecteur). Il faut donc calculer la
        # # matrice de transformation pour un axe donnée. C'est la liste
        # # de transformations qui sont entre cet axe et le detecteur.
        # axis_delta = None
        # axis_gamma = None

        # # il nous faut ensuite calculer la normale du plan dans lequel
        # # nous allons projeter les pixels. (C'est le produit vectoriel
        # # de k0, axis_xxx).
        # n_delta = None
        # n_gamma = None

        # # On calcule la projection sur la normale des plans en
        # # question.
        # p_delta = None
        # p_gamma = None

        # # On calcule la norme de chaque pixel. (qui pourra etre
        # # calcule une seule fois pour toutes les images).
        # l2 = numpy.linalg.norm(pixels, order=2, axis=-1)

        # # xxx0 is the angles of the diffractometer for the given
        # # image.
        # delta = numpy.arcsin(p_delta / l2) + delta0
        # gamma = numpy.arcsin(p_gamma / l2) + gamma0
        # omega = numpy.ones_like(delta) * omega0

        # return (omega, delta, gamma)

    def get_axis_labels(self) -> Tuple[str]:
        return 'delta', 'gamma', 'omega'


class AnglesProjection2(backend.ProjectionBase):    # omega <> mu
    def project(self, index: int, pdataframe: PDataFrame) -> Tuple[ndarray]:
        # put the detector at the right position

        pixels = pdataframe.pixels
        geometry = pdataframe.dataframe.diffractometer.geometry
        detrot = pdataframe.input_config.detrot
        sdd = pdataframe.input_config.sdd

        try:
            axis = geometry.axis_get("eta_a")
            eta_a = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            eta_a = 0
        try:
            axis = geometry.axis_get("mu")
            mu0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            mu0 = 0
        try:
            axis = geometry.axis_get("delta")
            delta0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            delta0 = 0
        try:
            axis = geometry.axis_get("gamma")
            gamma0 = axis.value_get(Hkl.UnitEnum.USER)
        except GLib.GError as err:
            gamma0 = 0

        P = M(math.radians(eta_a), [1, 0, 0])
        if detrot is not None:
            P = numpy.dot(P, M(math.radians(detrot), [1, 0, 0]))

        x, y, z = numpy.tensordot(P, pixels, axes=1)

        delta = numpy.rad2deg(numpy.arctan(z / sdd)) + delta0
        gamma = numpy.rad2deg(numpy.arctan(y / sdd)) + gamma0
        mu = numpy.ones_like(delta) * mu0

        return (delta, gamma, mu)

    def get_axis_labels(self) -> Tuple[str]:
        return 'delta', 'gamma', 'mu'

##################
# Input Backends #
##################


class SIXS(backend.InputBase):
    # OFFICIAL API

    dbg_scanno = None
    dbg_pointno = None

    def generate_jobs(self, command):
        scans = util.parse_multi_range(",".join(command).replace(" ", ","))
        if not len(scans):
            sys.stderr.write("error: no scans selected, nothing to do\n")
        for scanno in scans:
            util.status("processing scan {0}...".format(scanno))
            if self.config.pr:
                pointcount = self.config.pr[1] - self.config.pr[0] + 1
                start = self.config.pr[0]
            else:
                start = 0
                pointcount = self.get_pointcount(scanno)
            if pointcount > self.config.target_weight * 1.4:
                for s in util.chunk_slicer(pointcount, self.config.target_weight):
                    yield backend.Job(
                        scan=scanno,
                        firstpoint=start + s.start,
                        lastpoint=start + s.stop - 1,
                        weight=s.stop - s.start,
                    )
            else:
                yield backend.Job(
                    scan=scanno,
                    firstpoint=start,
                    lastpoint=start + pointcount - 1,
                    weight=pointcount,
                )

    def process_job(self, job):
        super(SIXS, self).process_job(job)
        with File(self.get_filename(job.scan), "r") as scan:
            self.metadict = dict()
            try:
                for dataframe in dataframes(scan, self.HPATH, self.config):
                    pixels = self.get_pixels(dataframe.detector)
                    mask = self.get_mask(dataframe.detector, self.config.maskmatrix)

                    for index in range(job.firstpoint, job.lastpoint + 1):
                        res = self.process_image(index, dataframe, pixels, mask)
                        # some frame could be skipped
                        if res is None:
                            util.status(f"skipped {index}")
                            continue
                        else:
                            yield res
                util.statuseol()
            except Exception as exc:
                exc.args = errors.addmessage(
                    exc.args,
                    ", An error occured for scan {0} at point {1}. See above for more information".format(
                        self.dbg_scanno, self.dbg_pointno
                    ),
                )  # noqa
                raise
            self.metadata.add_section("sixs_backend", self.metadict)

    def parse_config(self, config):
        super(SIXS, self).parse_config(config)
        # Optional, select a subset of the image range in the x
        # direction. all by default
        self.config.xmask = util.parse_multi_range(config.pop("xmask", None))

        # Optional, select a subset of the image range in the y
        # direction. all by default
        self.config.ymask = util.parse_multi_range(config.pop("ymask", None))

        # location of the nexus files (take precedence on nexusfile)
        self.config.nexusdir = config.pop("nexusdir", None)

        # Location of the specfile
        self.config.nexusfile = config.pop("nexusfile", None)

        # Optional, all range by default
        self.config.pr = config.pop("pr", None)
        if self.config.xmask is None:
            self.config.xmask = slice(None)
        if self.config.ymask is None:
            self.config.ymask = slice(None)
        if self.config.pr:
            self.config.pr = util.parse_tuple(
                self.config.pr, length=2, type=int
            )  # noqa

        # sample to detector distance (mm)
        self.config.sdd = float(config.pop("sdd"))

        # x,y coordinates of the central pixel
        self.config.centralpixel = util.parse_tuple(
            config.pop("centralpixel"), length=2, type=int
        )  # noqa

        # Optional, if supplied pixels where the mask is 0 will be removed
        self.config.maskmatrix = config.pop("maskmatrix", None)

        # detector rotation around x (1, 0, 0)
        self.config.detrot = config.pop("detrot", None)
        if self.config.detrot is not None:
            try:
                self.config.detrot = float(self.config.detrot)
            except ValueError:
                self.config.detrot = None

        # attenuation_coefficient (Optional)
        attenuation_coefficient = config.pop("attenuation_coefficient", None)
        if attenuation_coefficient is not None:
            try:
                self.config.attenuation_coefficient = float(
                    attenuation_coefficient
                )  # noqa
            except ValueError:
                self.config.attenuation_coefficient = None
        else:
            self.config.attenuation_coefficient = None

        # surface_orientation
        surface_orientation = config.pop("surface_orientation", None)
        surface_orientation_opt = SurfaceOrientation.VERTICAL
        if surface_orientation is not None:
            if surface_orientation.lower() == "horizontal":
                surface_orientation_opt = SurfaceOrientation.HORIZONTAL
        self.config.surface_orientation = surface_orientation_opt

        # sample
        self.config.a = util.parse_float(config, "a", None)
        self.config.b = util.parse_float(config, "b", None)
        self.config.c = util.parse_float(config, "c", None)
        self.config.alpha = util.parse_float(config, "alpha", None)
        self.config.beta = util.parse_float(config, "beta", None)
        self.config.gamma = util.parse_float(config, "gamma", None)
        self.config.ux = util.parse_float(config, "ux", None)
        self.config.uy = util.parse_float(config, "uy", None)
        self.config.uz = util.parse_float(config, "uz", None)

    def get_destination_options(self, command):
        if not command:
            return False
        command = ",".join(command).replace(" ", ",")
        scans = util.parse_multi_range(command)
        return dict(
            first=min(scans),
            last=max(scans),
            range=",".join(str(scan) for scan in scans),
        )  # noqa

    # CONVENIENCE FUNCTIONS
    def get_filename(self, scanno):
        filename = None
        if self.config.nexusdir:
            dirname = self.config.nexusdir
            files = [
                f
                for f in os.listdir(dirname)
                if (
                    (str(scanno).zfill(5) in f)
                    and (os.path.splitext(f)[1] in [".hdf5", ".nxs"])
                )
            ]
            if files is not []:
                filename = os.path.join(dirname, files[0])
        else:
            filename = self.config.nexusfile.format(scanno=str(scanno).zfill(5))  # noqa
        if not os.path.exists(filename):
            raise errors.ConfigError(
                "nexus filename does not exist: {0}".format(filename)
            )  # noqa
        return filename

    @staticmethod
    def apply_mask(data, xmask, ymask):
        roi = data[ymask, :]
        return roi[:, xmask]


class FlyScanUHV(SIXS):
    HPATH = {
        "image": HItem("xpad_image", False),
        "mu": HItem("UHV_MU", False),
        "omega": HItem("UHV_OMEGA", False),
        "delta": HItem("UHV_DELTA", False),
        "gamma": HItem("UHV_GAMMA", False),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }

    def get_pointcount(self, scanno):
        # just open the file in order to extract the number of step
        with File(self.get_filename(scanno), "r") as scan:
            return get_dataset(scan, self.HPATH["image"]).shape[0]

    def get_attenuation(self, index, h5_nodes, offset):
        attenuation = None
        if self.config.attenuation_coefficient is not None:
            try:
                try:
                    node = h5_nodes["attenuation"]
                    if node is not None:
                        attenuation = node[index + offset]
                    else:
                        raise Exception(
                            "you asked for attenuation but the file does not contain attenuation informations."
                        )  # noqa
                except KeyError:
                    attenuation = 1.0
            except ValueError:
                attenuation = WRONG_ATTENUATION
        return attenuation

    def get_timestamp(self, index, h5_nodes):
        timestamp = index
        if "timestamp" in h5_nodes:
            node = h5_nodes["timestamp"]
            if node is not None:
                timestamp = node[index]
        return timestamp

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        mu = h5_nodes["mu"][index]
        omega = h5_nodes["omega"][index]
        delta = h5_nodes["delta"][index]
        gamma = h5_nodes["gamma"][index]
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (mu, omega, delta, gamma))

    def process_image(
        self, index, dataframe, pixels, mask
    ) -> Optional[Tuple[ndarray, ndarray, Tuple[int, PDataFrame]]]:
        util.status(str(index))

        # extract the data from the h5 nodes

        h5_nodes = dataframe.h5_nodes
        intensity, attenuation, timestamp, values = self.get_values(index, h5_nodes)

        # the next version of the Hkl library will raise an exception
        # if at least one of the values is Nan/-Inf or +Inf. Emulate
        # this until we backported the right hkl library.
        if not all([math.isfinite(v) for v in values]):
            return None

        if attenuation is not None:
            if not math.isfinite(attenuation):
                return None

        # BEWARE in order to avoid precision problem we convert the
        # uint16 -> float32. (the size of the mantis is on 23 bits)
        # enought to contain the uint16. If one day we use uint32, it
        # should be necessary to convert into float64.
        intensity = intensity.astype("float32")

        weights = None
        if self.config.attenuation_coefficient is not None:
            if attenuation != WRONG_ATTENUATION:
                intensity *= self.config.attenuation_coefficient ** attenuation
                weights = numpy.ones_like(intensity)
                weights *= ~mask
            else:
                weights = numpy.zeros_like(intensity)
        else:
            weights = numpy.ones_like(intensity)
            weights *= ~mask

        k = 2 * math.pi / dataframe.source.wavelength

        hkl_geometry = dataframe.diffractometer.geometry
        hkl_geometry.axis_values_set(values, Hkl.UnitEnum.USER)

        # sample
        hkl_sample = dataframe.sample.sample
        q_sample = hkl_geometry.sample_rotation_get(hkl_sample)
        R = hkl_matrix_to_numpy(q_sample.to_matrix())

        # detector
        hkl_detector = dataframe.detector.detector
        q_detector = hkl_geometry.detector_rotation_get(hkl_detector)
        P = hkl_matrix_to_numpy(q_detector.to_matrix())

        if self.config.detrot is not None:
            P = numpy.dot(P, M(math.radians(self.config.detrot), [1, 0, 0]))

        surface_orientation = self.config.surface_orientation

        pdataframe = PDataFrame(
            pixels, k, dataframe.sample.ub, R, P, index, timestamp, surface_orientation, dataframe, self.config
        )

        return intensity, weights, (index, pdataframe)

    def get_pixels(self, detector):
        # works only for flat detector.
        if detector.name == "ufxc":
            max_shape = (256, 257)
            detector = pyFAI.detectors.Detector(75e-6, 75e-6, splineFile=None, max_shape=max_shape)
        else:
            detector = ALL_DETECTORS[detector.name]()

        y, x, _ = detector.calc_cartesian_positions()
        y0 = y[self.config.centralpixel[1], self.config.centralpixel[0]]
        x0 = x[self.config.centralpixel[1], self.config.centralpixel[0]]
        z = numpy.ones(x.shape) * -1 * self.config.sdd
        # return converted to the hkl library coordinates
        # x -> -y
        # y -> z
        # z -> -x
        return numpy.array([-z, -(x - x0), (y - y0)])

    def get_mask(self, detector: Detector, fnmask: Optional[str]=None) -> ndarray:
        if detector.name == "ufxc":
            mask = numpy.zeros((256, 257)).astype(bool)
        else:
            detector = ALL_DETECTORS[detector.name]()
            mask = detector.mask.astype(numpy.bool)
        maskmatrix = load_matrix(fnmask)
        if maskmatrix is not None:
            mask = numpy.bitwise_or(mask, maskmatrix)

        return mask


class FlyScanUHV2(FlyScanUHV):
    HPATH = {
        "image": HItem("xpad_image", False),
        "mu": HItem("mu", False),
        "omega": HItem("omega", False),
        "delta": HItem("delta", False),
        "gamma": HItem("gamma", False),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }

class FlyScanUHVUfxc(FlyScanUHV):
    HPATH = {
        "image": HItem("ufxc_sixs_image", False),
        "mu": HItem("mu", False),
        "omega": HItem("omega", False),
        "delta": HItem("delta", False),
        "gamma": HItem("gamma", False),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }


class FlyMedH(FlyScanUHV):
    HPATH = {
        "image": HItem("xpad_image", False),
        "pitch": HItem("beta", True),
        "mu": HItem("mu", False),
        "gamma": HItem("gamma", False),
        "delta": HItem("delta", False),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        pitch = h5_nodes["pitch"][index] if h5_nodes["pitch"] else 0.3
        mu = h5_nodes["mu"][index]
        gamma = h5_nodes["gamma"][index]
        delta = h5_nodes["delta"][index]
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (pitch, mu, gamma, delta))

class SBSMedH(FlyScanUHV):
    HPATH = {
        "image": DatasetPathWithAttribute("long_name", b"i14-c-c00/dt/xpad.1/image"),
        "pitch": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/diff-med-tpp/pitch"
        ),
        "mu": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-h-dif-group.1/mu"
        ),
        "gamma": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-h-dif-group.1/gamma"
        ),
        "delta": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-h-dif-group.1/delta"
        ),
        "attenuation": DatasetPathWithAttribute("long_name", b"i14-c-c00/ex/roic/att"),
        "timestamp": HItem("sensors_timestamps", True),
    }

    def get_pointcount(self, scanno: int) -> int:
        # just open the file in order to extract the number of step
        with File(self.get_filename(scanno), "r") as scan:
            path = self.HPATH["image"]
            return get_dataset(scan, path).shape[0]

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        pitch = h5_nodes["pitch"][index]
        mu = h5_nodes["mu"][index]
        gamma = h5_nodes["gamma"][index]
        delta = h5_nodes["delta"][index]
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (pitch, mu, gamma, delta))


class SBSFixedDetector(FlyScanUHV):
    HPATH = {
        "image": HItem("data_11", False),
        "timestamp": HItem("sensors_timestamps", True),
    }

    def get_pointcount(self, scanno):
        # just open the file in order to extract the number of step
        with File(self.get_filename(scanno), "r") as scan:
            return get_nxclass(scan, "NXdata")["data_11"].shape[0]

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, None)

    def process_image(
        self, index, dataframe, pixels, mask
    ) -> Optional[Tuple[ndarray, ndarray, Tuple[int, PDataFrame]]]:
        util.status(str(index))

        # extract the data from the h5 nodes

        h5_nodes = dataframe.h5_nodes
        intensity, attenuation, timestamp, _values = self.get_values(index, h5_nodes)

        if not math.isfinite(attenuation):
            return None

        # BEWARE in order to avoid precision problem we convert the
        # uint16 -> float32. (the size of the mantis is on 23 bits)
        # enought to contain the uint16. If one day we use uint32, it
        # should be necessary to convert into float64.
        intensity = intensity.astype("float32")

        weights = None
        if self.config.attenuation_coefficient is not None:
            if attenuation != WRONG_ATTENUATION:
                intensity *= self.config.attenuation_coefficient ** attenuation
                weights = numpy.ones_like(intensity)
                weights *= ~mask
            else:
                weights = numpy.zeros_like(intensity)
        else:
            weights = numpy.ones_like(intensity)
            weights *= ~mask

        k = 2 * math.pi / dataframe.source.wavelength

        I = numpy.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])

        if self.config.detrot is not None:
            P = M(math.radians(self.config.detrot), [1, 0, 0])

        pdataframe = PDataFrame(pixels, k, I, I, P, index, timestamp, dataframe, self.config)

        return intensity, weights, (index, pdataframe)


class FlyMedV(FlyScanUHV):
    HPATH = {
        "image": HItem("xpad_image", False),
        "beta": HItem("beta", True),
        "mu": HItem("mu", False),
        "omega": HItem("omega", False),
        "gamma": HItem("gamma", False),
        "delta": HItem("delta", False),
        "etaa": HItem("etaa", True),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        beta = h5_nodes["beta"][index] if h5_nodes["beta"] else 0.0
        mu = h5_nodes["mu"][index]
        omega = h5_nodes["omega"][index]
        gamma = h5_nodes["gamma"][index]
        delta = h5_nodes["delta"][index]
        etaa = h5_nodes["etaa"][index] if h5_nodes["etaa"] else 0.0
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (beta, mu, omega, gamma, delta, etaa))

class FlyMedVS70(FlyMedV):
    HPATH = {
        "image": HItem("xpad_s70_image", False),
        "beta": HItem("beta", True),
        "mu": HItem("mu", False),
        "omega": HItem("omega", False),
        "gamma": HItem("gamma", False),
        "delta": HItem("delta", False),
        "etaa": HItem("etaa", True),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
    }

class FLYMedVEiger(FlyMedV):
    HPATH = {
        "image": HItem("eiger_image", False),
        "beta": HItem("beta", True),
        "mu": HItem("mu", False),
        "omega": HItem("omega", False),
        "gamma": HItem("gamma", False),
        "delta": HItem("delta", False),
        "etaa": HItem("etaa", True),
        "attenuation": HItem("attenuation", True),
        "timestamp": HItem("epoch", True),
        "eix": DatasetPathOr( HItem("eix", True),
                              DatasetPathContains("i14-c-cx1-dt-det_tx.1/position_pre")),
        "eiz": DatasetPathOr( HItem("eiz", True),
                              DatasetPathContains("i14-c-cx1-dt-det_tz.1/position_pre"))
    }

    def get_translation(self, node, index, default):
        res = default
        if node:
            if node.shape[0] == 1:
                res = node[0]
            else:
                res = node[index]
        return res

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        beta = h5_nodes["beta"][index] if h5_nodes["beta"] else 0.0  # degrees
        mu = h5_nodes["mu"][index]  # degrees
        omega = h5_nodes["omega"][index]  # degrees
        gamma = 0  # degrees
        delta = 0  # degrees
        etaa = 0  # degrees
        eix = self.get_translation(h5_nodes["eix"], index, 0.0)  # mm
        eiz = self.get_translation(h5_nodes["eiz"], index, 0.0)  # mm
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, eix, eiz, (beta, mu, omega, gamma, delta, etaa))

    def process_image(
        self, index, dataframe, pixels0, mask
    ) -> Optional[Tuple[ndarray, ndarray, Tuple[int, PDataFrame]]]:
        util.status(str(index))

        # extract the data from the h5 nodes

        h5_nodes = dataframe.h5_nodes
        intensity, attenuation, timestamp, eix, eiz, values = self.get_values(index, h5_nodes)

        # TODO translate the detector, must be done after the detrot.
        pixels = pixels0.copy()
        if eix != 0.0:
            pixels[2] += eix * 1e-3
        if eiz != 0.0:
            pixels[1] += eiz * 1e-3

        # the next version of the Hkl library will raise an exception
        # if at least one of the values is Nan/-Inf or +Inf. Emulate
        # this until we backported the right hkl library.
        if not all([math.isfinite(v) for v in values]):
            return None

        if attenuation is not None:
            if not math.isfinite(attenuation):
                return None

        # BEWARE in order to avoid precision problem we convert the
        # uint16 -> float32. (the size of the mantis is on 23 bits)
        # enought to contain the uint16. If one day we use uint32, it
        # should be necessary to convert into float64.
        intensity = intensity.astype("float32")

        weights = None
        if self.config.attenuation_coefficient is not None:
            if attenuation != WRONG_ATTENUATION:
                intensity *= self.config.attenuation_coefficient ** attenuation
                weights = numpy.ones_like(intensity)
                weights *= ~mask
            else:
                weights = numpy.zeros_like(intensity)
        else:
            weights = numpy.ones_like(intensity)
            weights *= ~mask

        k = 2 * math.pi / dataframe.source.wavelength

        hkl_geometry = dataframe.diffractometer.geometry
        hkl_geometry.axis_values_set(values, Hkl.UnitEnum.USER)

        # sample
        hkl_sample = dataframe.sample.sample
        q_sample = hkl_geometry.sample_rotation_get(hkl_sample)
        R = hkl_matrix_to_numpy(q_sample.to_matrix())

        # detector
        hkl_detector = dataframe.detector.detector
        q_detector = hkl_geometry.detector_rotation_get(hkl_detector)
        P = hkl_matrix_to_numpy(q_detector.to_matrix())

        if self.config.detrot is not None:
            P = numpy.dot(P, M(math.radians(self.config.detrot), [1, 0, 0]))

        surface_orientation = self.config.surface_orientation

        pdataframe = PDataFrame(
            pixels, k, dataframe.sample.ub, R, P, index, timestamp, surface_orientation, dataframe, self.config
        )

        return intensity, weights, (index, pdataframe)


class SBSMedV(FlyScanUHV):
    HPATH = {
        "image": DatasetPathWithAttribute("long_name", b"i14-c-c00/dt/xpad.1/image"),
        "beta": DatasetPathContains("i14-c-cx1-ex-diff-med-tpp/TPP/Orientation/pitch"),
        "mu": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/mu"
        ),
        "omega": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/omega"
        ),
        "gamma": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/gamma"
        ),
        "delta": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/delta"
        ),
        "etaa": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/etaa"
        ),
        "attenuation": DatasetPathWithAttribute("long_name", b"i14-c-c00/ex/roic/att"),
        "timestamp": HItem("sensors_timestamps", True),
    }

    def get_pointcount(self, scanno: int) -> int:
        # just open the file in order to extract the number of step
        with File(self.get_filename(scanno), "r") as scan:
            path = self.HPATH["image"]
            return get_dataset(scan, path).shape[0]

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        beta = h5_nodes["beta"][0]
        mu = h5_nodes["mu"][index]
        omega = h5_nodes["omega"][index]
        gamma = h5_nodes["gamma"][index]
        delta = h5_nodes["delta"][index]
        etaa = h5_nodes["etaa"][index]
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (beta, mu, omega, gamma, delta, etaa))

class SBSMedVFixDetector(SBSMedV):
    HPATH = {
        "image": DatasetPathWithAttribute("long_name", b"i14-c-c00/dt/eiger.1/image"),
        "beta": DatasetPathContains("i14-c-cx1-ex-diff-med-tpp/TPP/Orientation/pitch"),
        "mu": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/mu"
        ),
        "omega": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/omega"
        ),
        "gamma": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/gamma"
        ),
        "delta": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/delta"
        ),
        "etaa": DatasetPathWithAttribute(
            "long_name", b"i14-c-cx1/ex/med-v-dif-group.1/etaa"
        ),
        "attenuation": DatasetPathWithAttribute("long_name", b"i14-c-c00/ex/roic/att"),
        "timestamp": HItem("sensors_timestamps", True),
    }

    def get_values(self, index, h5_nodes):
        image = h5_nodes["image"][index]
        beta = h5_nodes["beta"][0]
        mu = h5_nodes["mu"][index]
        omega = h5_nodes["omega"][index]
        gamma = 0
        delta = 0
        etaa = 0
        attenuation = self.get_attenuation(index, h5_nodes, 2)
        timestamp = self.get_timestamp(index, h5_nodes)

        return (image, attenuation, timestamp, (beta, mu, omega, gamma, delta, etaa))

def load_matrix(filename):
    if filename is None:
        return None
    if os.path.exists(filename):
        ext = os.path.splitext(filename)[-1]
        if ext == ".txt":
            return numpy.array(numpy.loadtxt(filename), dtype=numpy.bool)
        elif ext == ".npy":
            mask = numpy.array(numpy.load(filename), dtype=numpy.bool)
            print("loaded mask sum: ", numpy.sum(mask))
            return mask
        else:
            raise ValueError(
                "unknown extension {0}, unable to load matrix!\n".format(ext)
            )  # noqa
    else:
        raise IOError(
            "filename: {0} does not exist. Can not load matrix".format(filename)
        )  # noqa



"""
BINocular backend for beamline ID03:EH2 
This backend should serve as a basic example of a backend based on
xrayutilities [1]. It still uses PyMCA for parsing the spec,edf files.
The 'original' ID03 backend was used as a template.

Created on 2014-10-16

[1] http://xrayutilities.sourceforge.net/

author: Dominik Kriegner (dominik.kriegner@gmail.com)
"""

import sys
import os
import glob
import numpy

import xrayutilities as xu

from PyMca5.PyMca import EdfFile, specfile, specfilewrapper

from .. import backend, errors, util


class HKLProjection(backend.ProjectionBase):
    # scalars: mu, theta, [chi, phi, "omitted"] delta, gamR, gamT, ty, wavelength
    # 3x3 matrix: UB
    def project(self, mu, theta, delta, gamR, gamT, ty, wavelength, UB, qconv):
        qconv.wavelength = wavelength
        h, k, l = qconv.area(
            mu, theta, mu, delta, ty, gamT, gamR, UB=UB.reshape((3, 3))
        )
        return (h, k, l)

    def get_axis_labels(self):
        return "H", "K", "L"


class HKProjection(HKLProjection):
    def project(self, mu, theta, delta, gamR, gamT, ty, wavelength, UB, qconv):
        H, K, L = super(HKProjection, self).project(
            mu, theta, delta, gamR, gamT, ty, wavelength, UB, qconv
        )
        return (H, K)

    def get_axis_labels(self):
        return "H", "K"


class QProjection(backend.ProjectionBase):
    def project(self, mu, theta, delta, gamR, gamT, ty, wavelength, UB, qconv):
        qconv.wavelength = wavelength
        qx, qy, qz = qconv.area(
            mu, theta, mu, delta, ty, gamT, gamR, UB=numpy.identity(3)
        )
        return (qx, qy, qz)

    def get_axis_labels(self):
        return "qx", "qy", "qz"


class ID03Input(backend.InputBase):
    # OFFICIAL API
    def generate_jobs(self, command):
        scans = util.parse_multi_range(",".join(command).replace(" ", ","))
        if not len(scans):
            sys.stderr.write("error: no scans selected, nothing to do\n")
        for scanno in scans:
            scan = self.get_scan(scanno)
            try:
                pointcount = scan.lines()
            except specfile.error:  # no points
                continue
            next(self.get_images(scan, 0, pointcount - 1, dry_run=True))  # dryrun

            if (
                self.config.target_weight
                and pointcount > self.config.target_weight * 1.4
            ):
                for s in util.chunk_slicer(pointcount, self.config.target_weight):
                    yield backend.Job(
                        scan=scanno,
                        firstpoint=s.start,
                        lastpoint=s.stop - 1,
                        weight=s.stop - s.start,
                    )
            else:
                yield backend.Job(
                    scan=scanno,
                    firstpoint=0,
                    lastpoint=pointcount - 1,
                    weight=pointcount,
                )

    def process_job(self, job):
        super(ID03Input, self).process_job(job)
        scan = self.get_scan(job.scan)

        scanparams = self.get_scan_params(scan)  # wavelength, UB
        pointparams = self.get_point_params(
            scan, job.firstpoint, job.lastpoint
        )  # 1D array of diffractometer angles + mon + transm
        images = self.get_images(scan, job.firstpoint, job.lastpoint)  # iterator!

        for pp, image in zip(pointparams, images):
            yield self.process_image(scanparams, pp, image)

    def parse_config(self, config):
        super(ID03Input, self).parse_config(config)
        self.config.xmask = util.parse_multi_range(config.pop("xmask"))
        self.config.ymask = util.parse_multi_range(config.pop("ymask"))
        self.config.specfile = config.pop("specfile")
        self.config.imagefolder = config.pop("imagefolder", None)
        self.config.UB = config.pop("ub", None)
        if self.config.UB:
            self.config.UB = util.parse_tuple(self.config.UB, length=9, type=float)
        self.config.sdd = float(config.pop("sdd"))
        self.config.pixelsize = util.parse_tuple(
            config.pop("pixelsize"), length=2, type=float
        )
        self.config.centralpixel = util.parse_tuple(
            config.pop("centralpixel"), length=2, type=int
        )

    def get_destination_options(self, command):
        if not command:
            return False
        command = ",".join(command).replace(" ", ",")
        scans = util.parse_multi_range(command)
        return dict(first=min(scans), last=max(scans), range=",".join(command))

    # CONVENIENCE FUNCTIONS
    _spec = None

    def get_scan(self, scannumber):
        if self._spec is None:
            self._spec = specfilewrapper.Specfile(self.config.specfile)
        return self._spec.select("{0}.1".format(scannumber))

    def find_edfs(self, pattern, scanno):
        files = glob.glob(pattern)
        ret = {}
        for file in files:
            try:
                filename = os.path.basename(file).split(".")[0]
                scan, point, image = filename.split("_")[-3:]
                scan, point, image = int(scan), int(point), int(image)
                if scan == scanno and point not in list(ret.keys()):
                    ret[point] = file
            except ValueError:
                continue
        return ret

    @staticmethod
    def apply_mask(data, xmask, ymask):
        roi = data[ymask, :]
        return roi[:, xmask]

    # MAIN LOGIC
    def get_scan_params(self, scan):
        UB = numpy.array(scan.header("G")[2].split(" ")[-9:], dtype=numpy.float)
        wavelength = float(scan.header("G")[1].split(" ")[-1])

        return wavelength, UB

    def get_images(self, scan, first, last, dry_run=False):
        try:
            uccdtagline = scan.header("UCCD")[0]
            UCCD = os.path.split(os.path.dirname(uccdtagline.split()[-1]))
        except:
            print(
                "warning: UCCD tag not found, use imagefolder for proper file specification"
            )
            UCCD = []
        pattern = self._get_pattern(UCCD)
        matches = self.find_edfs(pattern, scan.number())
        if set(range(first, last + 1)) > set(matches.keys()):
            raise errors.FileError(
                "incorrect number of matches for scan {0} using pattern {1}".format(
                    scan.number(), pattern
                )
            )
        if dry_run:
            yield
        else:
            for i in range(first, last + 1):
                edf = EdfFile.EdfFile(matches[i])
                yield edf.GetData(0)

    def _get_pattern(self, UCCD):
        imagefolder = self.config.imagefolder
        if imagefolder:
            try:
                imagefolder = imagefolder.format(UCCD=UCCD, rUCCD=list(reversed(UCCD)))
            except Exception as e:
                raise errors.ConfigError(
                    "invalid 'imagefolder' specification '{0}': {1}".format(
                        self.config.imagefolder, e
                    )
                )
        else:
            imagefolder = os.path.join(*UCCD)

        if not os.path.exists(imagefolder):
            raise ValueError(
                "invalid 'imagefolder' specification '{0}'. Path {1} does not exist".format(
                    self.config.imagefolder, imagefolder
                )
            )
        return os.path.join(imagefolder, "*")


class EH2(ID03Input):
    monitor_counter = "Monitor"
    # define ID03 goniometer, SIXC geometry with 2D detector mounted on a
    # translation-axis (distance changing with changing Gamma)
    # The geometry is: 1+3S+2D
    # sample axis mu, th, chi, phi -> here chi,phi are omitted
    # detector axis mu, del, gam
    # gam is realized by a translation along z (gamT) and rotation around x+ (gamR)
    qconv = xu.experiment.QConversion(
        ["x+", "z-"], ["x+", "z-", "ty", "tz", "x+"], [0, 1, 0]  # 'y+', 'z+'
    )
    # convention for coordinate system: y downstream; z outwards; x upwards
    # (righthanded)
    # QConversion will set up the goniometer geometry. So the first argument
    # describes the sample rotations, the second the detector rotations and the
    # third the primary beam direction.
    ty = 600.0  # mm

    def parse_config(self, config):
        super(EH2, self).parse_config(config)
        centralpixel = self.config.centralpixel  # (row, column) = (gamma, delta)
        # define detector parameters
        roi = (
            self.config.ymask[0],
            self.config.ymask[-1] + 1,
            self.config.xmask[0],
            self.config.xmask[-1] + 1,
        )
        self.qconv.init_area(
            "x+",
            "z-",
            cch1=centralpixel[1],
            cch2=centralpixel[0],
            Nch1=516,
            Nch2=516,
            pwidth1=self.config.pixelsize[1],
            pwidth2=self.config.pixelsize[0],
            distance=self.config.sdd - self.ty,
            roi=roi,
        )
        # distance sdd-600 corresponds to distance of the detector chip from
        # the gamR rotation axis (rest is handled by the translations ty and
        # gamT (along z))
        print(("{:>9} {:>10} {:>9} {:>9}".format("Mu", "Theta", "Delta", "Gamma")))

    def process_image(self, scanparams, pointparams, image):
        mu, theta, chi, phi, delta, gamma, mon, transm = pointparams
        wavelength, UB = scanparams
        data = image / mon / transm
        print(("{:9.4f} {:10.4f} {:9.4f} {:9.4f}".format(mu, theta, delta, gamma)))

        # recalculate detector translation (which should be saved!)
        gamT = self.ty * numpy.tan(numpy.radians(gamma))

        # masking
        intensity = self.apply_mask(data, self.config.xmask, self.config.ymask)

        # no polarization correction for the moment!

        return (
            intensity,
            numpy.ones_like(intensity),
            (
                mu,
                theta,
                delta,
                gamma,
                gamT,  # weights added to API. keeps functionality identical with wights of one
                self.ty,
                wavelength,
                UB,
                self.qconv,
            ),
        )

    def get_point_params(self, scan, first, last):
        sl = slice(first, last + 1)

        MU, TH, CHI, PHI, DEL, GAM, MON, TRANSM = list(range(8))
        params = numpy.zeros((last - first + 1, 8))
        # Mu, Theta, Chi, Phi, Delta, Gamma, MON, transm
        params[:, CHI] = scan.motorpos("Chi")
        params[:, PHI] = scan.motorpos("Phi")
        params[:, TH] = scan.datacol("thcnt")[sl]
        params[:, GAM] = scan.datacol("gamcnt")[sl]
        params[:, DEL] = scan.datacol("delcnt")[sl]
        params[:, MON] = scan.datacol(self.monitor_counter)[sl]
        params[:, TRANSM] = scan.datacol("transm")[sl]
        params[:, MU] = scan.datacol("mucnt")[sl]
        return params



import binoculars

# to load a binoculars space file in a python script use the load function. You can supply a key if you do not
# want to load the entire space. By default the space is loaded entirely.
space = binoculars.load("test.hdf5")

# if you quickly want to see what the content of the space is you can print the space. This will provide information about the
# dimension, the labels, and the range of the space, but not of the data in the space.
print space

# you can slice a space. This will return a new space with the limits set by the sliced values. The sliced values are given in terms of the
# labele values, not in terms of the indices of the space. You can use the same language as is used in slicing numpy arrays.
newspace = space[-0.01:0.01, -0.05:0.05, :]
# this is equivalent to
newspace = space.slice("qx", slice(-0.01, 0.01)).slice("qy", slice(-0.05, 0.05))

# You can project the space on its axes by supplying an arbitrary number of axes in the project function.
specular = newspace.project("qx", "qy")

# to explore the data inside a space you can call the get_masked() function. This will return you the data as an ndimensional masked array with
# the values masked where there are no datapoints.
data = space.get_masked()

# if you need the underlying cooridnates of the data you can ask extract the grid with
grid = space.get_grid()

# you can add two spaces provided they have the same labels and the same resolution. The range of the space can be arbitrary.
total = binoculars.load("test.hdf5") + binoculars.load("test1.hdf5")

# you can view the configuration settings used to create the space.
binoculars.info(space)

# if you now want to reuse this configuration file as an input you can extract the configuration file by
space.config.totxtfile("config.txt")

# You can save the space in another format by changing the extension, currently only txt and hdf5 supported, and EDF if you have PyMca installed
binoculars.save("test.txt", space)

# you can plot a space in a python script, or in an interactive terminal using the binoculars.plotspace function. This function
# automatically puts the right coordinates on the axes. Check the advanced options by typing 'help(binoculars.plotspace)'
import matplotlib.pyplot as pyplot

binoculars.plotspace(specular)
pyplot.show()



import numpy
import matplotlib.pyplot as pyplot
import binoculars
import binoculars.space

space = (
    binoculars.load("test.hdf5")
    .slice("delta", slice(-0.035, 0.035))
    .project("delta")
    .slice("g-m", slice(None, 0.4))
)

results = []
x = binoculars.space.get_axis_values(space, "g+m")

for index, curve in enumerate(binoculars.space.iterate_over_axis(space, "g+m")):
    try:
        fit = binoculars.fitspace(curve, "lorentzian")
        results.append(numpy.append(x[index], fit.result))
        # pyplot.figure()
        # binoculars.plotspace(curve, fit = fit.fitdata, log = False)
        # pyplot.savefig('fit_{0}'.format(index))
        # pyplot.close()
    except Exception as e:
        print "warning: fit {0} failed: {1}".format(index, e)

result = numpy.vstack(results)
numpy.savetxt("fitdata.txt", result)



import argparse
import binascii
import configparser
import contextlib
import copy
import glob
import gzip
import inspect
import io
import itertools
import json
import os
import pickle
import random
import re
import struct
import socket
import sys
import time

import h5py
import numpy

from PyMca5.PyMca import EdfFile

from . import errors

### ARGUMENT HANDLING


def as_string(text):
    if hasattr(text, "decode"):
        text = text.decode()
    return text


class OrderedOperation(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        oops = getattr(namespace, "ordered_operations", [])
        oops.append((self.dest, values))
        setattr(namespace, "ordered_operations", oops)


def argparse_common_arguments(parser, *args):
    for arg in args:
        # (ORDERED) OPERATIONS
        if arg == "project":
            parser.add_argument(
                "-p",
                "--project",
                metavar="AXIS",
                action=OrderedOperation,
                help="project space on AXIS",
            )
        elif arg == "slice":
            parser.add_argument(
                "--slice",
                nargs=2,
                metavar=("AXIS", "START:STOP"),
                action=OrderedOperation,
                help="slice AXIS from START to STOP (replace minus signs by 'm')",
            )
        elif arg == "pslice":
            parser.add_argument(
                "--pslice",
                nargs=2,
                metavar=("AXIS", "START:STOP"),
                action=OrderedOperation,
                help="like slice, but also project on AXIS after slicing",
            )
        elif arg == "transform":
            parser.add_argument(
                "--transform",
                metavar="VAR@RES=EXPR;VAR2@RES2=EXPR2;...",
                action=OrderedOperation,
                help="perform coordinate transformation, rebinning data on new axis named VAR with resolution RES defined by EXPR, example: Q@0.1=sqrt(H**2+K**2+L**2)",
            )
        elif arg == "rebin":
            parser.add_argument(
                "--rebin",
                metavar="N,M,...",
                action=OrderedOperation,
                help="reduce binsize by factor N in first dimension, M in second, etc",
            )

        # SUBTRACT
        elif arg == "subtract":
            parser.add_argument(
                "--subtract", metavar="SPACE", help="subtract SPACE from input file"
            )

        # PRESENTATION
        elif arg == "nolog":
            parser.add_argument(
                "--nolog", action="store_true", help="do not use logarithmic axis"
            )
        elif arg == "clip":
            parser.add_argument(
                "-c",
                "--clip",
                metavar="FRACTION",
                default=0.00,
                help="clip color scale to remove FRACTION datapoints",
            )

        # OUTPUT
        elif arg == "savepdf":
            parser.add_argument(
                "-s",
                "--savepdf",
                action="store_true",
                help="save output as pdf, automatic file naming",
            )
        elif arg == "savefile":
            parser.add_argument(
                "--savefile",
                metavar="FILENAME",
                help="save output as FILENAME, autodetect filetype",
            )

        # ERROR!
        else:
            raise ValueError("unsupported argument '{0}'".format(arg))


def parse_transform_args(transform):
    for t in transform.split(";"):
        lhs, expr = t.split("=")
        ax, res = lhs.split("@")
        yield ax.strip(), float(res), expr.strip()


def handle_ordered_operations(space, args, auto3to2=False):
    info = []
    for command, opts in getattr(args, "ordered_operations", []):

        if command == "slice" or command == "pslice":
            ax, key = opts
            axindex = space.axes.index(ax)
            axlabel = space.axes[axindex].label
            if ":" in key:
                start, stop = key.split(":")
                if start:
                    start = float(start.replace("m", "-"))
                else:
                    start = space.axes[axindex].min
                if stop:
                    stop = float(stop.replace("m", "-"))
                else:
                    stop = space.axes[axindex].max
                key = slice(start, stop)

                info.append(
                    "sliced in {0} from {1} to {2}".format(axlabel, start, stop)
                )
            else:
                key = float(key.replace("m", "-"))
                info.append("sliced in {0} at {1}".format(axlabel, key))
            space = space.slice(axindex, key)

            if command == "pslice":
                try:
                    projectaxis = space.axes.index(ax)
                except ValueError:
                    pass
                else:
                    info.append(
                        "projected on {0}".format(space.axes[projectaxis].label)
                    )
                    space = space.project(projectaxis)

        elif command == "project":
            projectaxis = space.axes.index(opts)
            info.append("projected on {0}".format(space.axes[projectaxis].label))
            space = space.project(projectaxis)

        elif command == "transform":
            labels, resolutions, exprs = list(zip(*parse_transform_args(opts)))
            transformation = transformation_from_expressions(space, exprs)
            info.append(
                "transformed to {0}".format(
                    ", ".join(
                        "{0} = {1}".format(label, expr)
                        for (label, expr) in zip(labels, exprs)
                    )
                )
            )
            space = space.transform_coordinates(resolutions, labels, transformation)

        elif command == "rebin":
            if "," in opts:
                factors = tuple(int(i) for i in opts.split(","))
            else:
                factors = (int(opts),) * space.dimension
            space = space.rebin(factors)

        else:
            raise ValueError("unsported Ordered Operation '{0}'".format(command))

    if auto3to2 and space.dimension == 3:  # automatic projection on smallest axis
        projectaxis = numpy.argmin(space.photons.shape)
        info.append("projected on {0}".format(space.axes[projectaxis].label))
        space = space.project(projectaxis)

    return space, info


### STATUS LINES

_status_line_length = 0


def status(line, eol=False):
    """Prints a status line to sys.stdout, overwriting the previous one.
    Set eol to True to append a newline to the end of the line"""

    global _status_line_length
    sys.stdout.write("\r{0}\r{1}".format(" " * _status_line_length, line))
    if eol:
        sys.stdout.write("\n")
        _status_line_length = 0
    else:
        _status_line_length = len(line)

    sys.stdout.flush()


def statusnl(line):
    """Shortcut for status(..., eol=True)"""
    return status(line, eol=True)


def statuseol():
    """Starts a new status line, keeping the previous one intact"""
    global _status_line_length
    _status_line_length = 0
    sys.stdout.write("\n")
    sys.stdout.flush()


def statuscl():
    """Clears the status line, shortcut for status('')"""
    return status("")


### Dispatcher, projection and input finder
def get_backends():
    modules = glob.glob(os.path.join(os.path.dirname(__file__), "backends", "*.py"))
    names = list()

    for module in modules:
        if not module.endswith("__init__.py"):
            names.append(os.path.splitext(os.path.basename(module))[0])
    return names


def get_projections(module):
    from . import backend

    return get_base(module, backend.ProjectionBase)


def get_inputs(module):
    from . import backend

    return get_base(module, backend.InputBase)


def get_dispatchers():
    from . import dispatcher
    from inspect import isclass

    items = dir(dispatcher)

    options = []
    for item in items:
        obj = getattr(dispatcher, item)
        if isclass(obj):
            if issubclass(obj, dispatcher.DispatcherBase):
                options.append(item)

    return options


def get_base(modname, base):
    from inspect import isclass

    if modname not in get_backends():
        raise KeyError("{0} is not an available backend".format(modname))
    try:
        backends = __import__(
            "backends.{0}".format(modname), globals(), locals(), [], 1
        )
    except ImportError as e:
        raise ImportError(
            "Unable to import module backends.{0}: {1}".format(modname, e)
        )

    backend = getattr(backends, modname)
    items = dir(backend)

    options = []
    for item in items:
        obj = getattr(backend, item)
        if isclass(obj):
            if issubclass(obj, base):
                options.append(item)
    return options


### Dispatcher, projection and input configuration options finder


def get_dispatcher_configkeys(classname):
    from . import dispatcher

    cls = getattr(dispatcher, classname)
    return get_configkeys(cls)


def get_projection_configkeys(modname, classname):
    return get_backend_configkeys(modname, classname)


def get_input_configkeys(modname, classname):
    return get_backend_configkeys(modname, classname)


def get_backend_configkeys(modname, classname):
    backends = __import__("backends.{0}".format(modname), globals(), locals(), [], 1)
    backend = getattr(backends, modname)
    cls = getattr(backend, classname)
    return get_configkeys(cls)


def get_configkeys(cls):
    from inspect import getsource

    items = list()
    while hasattr(cls, "parse_config"):
        code = getsource(cls.parse_config)
        for line in code.split("\n"):
            key = parse_configcode(line)
            if key:
                if key not in items:
                    items.append(key)
        cls = cls.__base__
    return items


def parse_configcode(line):
    try:
        comment = "#".join(line.split("#")[1:])
        line = line.split("#")[0]
        index = line.index("config.pop")
        item = line[index:].split("'")[1]
        if item == "action":
            return  # action is reserved for internal use!
        return item, comment
    except ValueError:
        pass


### CONFIGURATION MANAGEMENT


def parse_float(config, option, default=None):
    value = default
    value_as_string = config.pop(option, None)
    if value_as_string is not None:
        try:
            value = float(value_as_string)  # noqa
        except ValueError:
            pass
    return value


def parse_range(r):
    if "-" in r:
        a, b = r.split("-")
        return list(range(int(a), int(b) + 1))
    elif r:
        return [int(r)]
    else:
        return []


def parse_multi_range(s):
    if not s:
        return s
    out = []
    ranges = s.split(",")
    for r in ranges:
        out.extend(parse_range(r))
    return out


def parse_tuple(s, length=None, type=str):
    if not s:
        return s
    t = tuple(type(i) for i in s.split(","))
    if length is not None and len(t) != length:
        raise ValueError(
            "invalid tuple length: expected {0} got {1}".format(length, len(t))
        )
    return t


def parse_bool(s):
    l = s.lower()
    if l in ("1", "true", "yes", "on"):
        return True
    elif l in ("0", "false", "no", "off"):
        return False
    raise ValueError("invalid input for boolean: '{0}'".format(s))


def parse_pairs(s):
    if not s:
        return s
    limits = []
    for lim in re.findall("\[(.*?)\]", s):
        parsed = []
        for pair in re.split(",", lim):
            mi, ma = tuple(m.strip() for m in pair.split(":"))
            if mi == "" and ma == "":
                parsed.append(slice(None))
            elif mi == "":
                parsed.append(slice(None, float(ma)))
            elif ma == "":
                parsed.append(slice(float(mi), None))
            else:
                if float(ma) < float(mi):
                    raise ValueError(
                        "invalid input. maximum is larger than minimum: '{0}'".format(s)
                    )
                else:
                    parsed.append(slice(float(mi), float(ma)))
        limits.append(parsed)
    return limits


def limit_to_filelabel(s):
    return tuple(
        "[{0}]".format(lim.replace("-", "m").replace(":", "-").replace(" ", ""))
        for lim in re.findall("\[(.*?)\]", s)
    )


class MetaBase(object):
    def __init__(self, label=None, section=None):
        self.sections = []
        if label is not None and section is not None:
            self.sections.append(label)
            setattr(self, label, section)
        elif label is not None:
            self.sections.append(label)
            setattr(self, label, dict())

    def add_section(self, label, section=None):
        self.sections.append(label)
        if section is not None:
            setattr(self, label, section)
        else:
            setattr(self, label, dict())

    def __repr__(self):
        str = "{0.__class__.__name__}{{\n".format(self)
        for section in self.sections:
            str += "  [{}]\n".format(section)
            s = getattr(self, section)
            for entry in s:
                str += "    {} = {}\n".format(entry, s[entry])
        str += "}\n"
        return str

    def copy(self):
        return copy.deepcopy(self)

    def serialize(self):
        sections = {}
        for section in self.sections:
            section_dict = {}
            attr = getattr(self, section)
            for key in list(attr.keys()):
                if isinstance(
                    attr[key], numpy.ndarray
                ):  # to be able to include numpy arrays in the serialisation
                    sio = io.StringIO()
                    numpy.save(sio, attr[key])
                    sio.seek(0)
                    section_dict[key] = binascii.b2a_hex(
                        sio.read()
                    )  # hex codation is needed to let json work with the string
                else:
                    section_dict[key] = attr[key]
            sections[section] = section_dict
        return json.dumps(sections)

    @classmethod
    def fromserial(cls, s):
        obj = cls()
        data = json.loads(s)
        for section in list(data.keys()):
            section_dict = data[section]
            for key in list(section_dict.keys()):
                if isinstance(
                    section_dict[key], str
                ):  # find and replace all the numpy serialised objects
                    if section_dict[key].startswith(
                        "934e554d505901004600"
                    ):  # numpy marker
                        sio = io.StringIO()
                        sio.write(binascii.a2b_hex(section_dict[key]))
                        sio.seek(0)
                        section_dict[key] = numpy.load(sio)
            setattr(obj, section, data[section])
            if section not in obj.sections:
                obj.sections.append(section)
        return obj


class MetaData(object):  # a collection of metadata objects
    def __init__(self):
        self.metas = []

    def add_dataset(self, dataset):
        if not isinstance(dataset, MetaBase) and not isinstance(dataset, ConfigFile):
            raise ValueError("MetaBase instance expected")
        else:
            self.metas.append(dataset)

    def __add__(self, other):
        new = self.__class__()
        new += self
        new += other
        return new

    def __iadd__(self, other):
        self.metas.extend(other.metas)
        return self

    @classmethod
    def fromfile(cls, filename):
        if isinstance(filename, str):
            if not os.path.exists(filename):
                raise IOError(
                    "Error importing configuration file. filename {0} does not exist".format(
                        filename
                    )
                )

        metadataobj = cls()
        with open_h5py(filename, "r") as fp:
            try:
                metadata = fp["metadata"]
            except KeyError:
                metadata = []  # when metadata is not present, proceed without Error
            for label in metadata:
                meta = MetaBase()
                for section in list(metadata[label].keys()):
                    group = metadata[label][section]
                    setattr(
                        meta, section, dict((key, group[key][()]) for key in group)
                    )
                    meta.sections.append(section)
                metadataobj.metas.append(meta)
        return metadataobj

    def tofile(self, filename):
        with open_h5py(filename, "w") as fp:
            metadata = fp.create_group("metadata")
            for meta in self.metas:
                label = find_unused_label("metasection", list(metadata.keys()))
                metabase = metadata.create_group(label)
                for section in meta.sections:
                    sectiongroup = metabase.create_group(section)
                    s = getattr(meta, section)
                    for key in list(s.keys()):
                        sectiongroup.create_dataset(key, data=s[key])

    def __repr__(self):
        str = "{0.__class__.__name__}{{\n".format(self)
        for meta in self.metas:
            for line in meta.__repr__().split("\n"):
                str += "    " + line + "\n"
        str += "}\n"
        return str

    def serialize(self):
        return json.dumps(list(meta.serialize() for meta in self.metas))

    @classmethod
    def fromserial(cls, s):
        obj = cls()
        for item in json.loads(s):
            obj.metas.append(MetaBase.fromserial(item))
        return obj


# Contains the unparsed config dicts


class ConfigFile(MetaBase):
    def __init__(self, origin="n/a", command=[]):
        self.origin = origin
        self.command = command
        super(ConfigFile, self).__init__()
        self.sections = ["dispatcher", "projection", "input"]
        for section in self.sections:
            setattr(self, section, dict())

    @classmethod
    def fromfile(cls, filename):
        if isinstance(filename, str):
            if not os.path.exists(filename):
                raise IOError(
                    "Error importing configuration file. filename {0} does not exist".format(
                        filename
                    )
                )

        configobj = cls(str(filename))
        with open_h5py(filename, "r") as fp:
            try:
                config = fp["configuration"]
                if "command" in config.attrs:
                    configobj.command = json.loads(as_string(config.attrs["command"]))
                for section in config:
                    if isinstance(config[section], h5py._hl.group.Group):  # new
                        setattr(
                            configobj,
                            section,
                            dict(
                                (key, config[section][key][()])
                                for key in config[section]
                            ),
                        )
                    else:  # old
                        setattr(configobj, section, dict(config[section]))
            except KeyError:
                pass  # when config is not present, proceed without Error
        return configobj

    @classmethod
    def fromtxtfile(cls, filename, command=[], overrides=[]):
        if not os.path.exists(filename):
            raise IOError(
                "Error importing configuration file. filename {0} does not exist".format(
                    filename
                )
            )

        config = configparser.RawConfigParser()
        config.read(filename)

        for section, option, value in overrides:
            config.set(section, option, value)

        configobj = cls(filename, command=command)
        for section in configobj.sections:
            setattr(
                configobj,
                section,
                dict((k, v.split("#")[0].strip()) for (k, v) in config.items(section)),
            )
        return configobj

    def tofile(self, filename):
        with open_h5py(filename, "w") as fp:
            conf = fp.create_group("configuration")
            conf.attrs["origin"] = str(self.origin)
            conf.attrs["command"] = json.dumps(self.command)
            for section in self.sections:
                sectiongroup = conf.create_group(section)
                s = getattr(self, section)
                for key in list(s.keys()):
                    sectiongroup.create_dataset(key, data=s[key])

    def totxtfile(self, filename):
        with open(filename, "w") as fp:
            fp.write("# Configurations origin: {}\n".format(self.origin))
            for section in self.sections:
                fp.write("[{}]\n".format(section))
                s = getattr(self, section)
                for entry in s:
                    fp.write("{} = {}\n".format(entry, s[entry]))

    def __repr__(self):
        str = super(ConfigFile, self).__repr__()
        str += "origin = {0}\n".format(self.origin)
        str += "command = {0}".format(",".join(self.command))
        return str


# contains one parsed dict, for distribution to dispatcher, input or projection class


class ConfigSection(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def copy(self):
        return copy.deepcopy(self)


# contains the parsed configsections


class ConfigSectionGroup(object):
    def __init__(self, origin="n/a"):
        self.origin = origin
        self.sections = "dispatcher", "projection", "input"
        for section in self.sections:
            setattr(self, section, ConfigSection())
        self.configfile = ConfigFile()


class ConfigurableObject(object):
    def __init__(self, config):
        if isinstance(config, ConfigSection):
            self.config = config
        elif not isinstance(config, dict):
            raise ValueError(
                "expecting dict or Configsection, not: {0}".format(type(config))
            )
        else:
            self.config = ConfigSection()
            try:
                allkeys = list(config.keys())
                self.parse_config(config)
            except KeyError as exc:
                raise errors.ConfigError(
                    "Configuration option {0} is missing from the configuration file. Please specify this option in the configuration file".format(
                        exc
                    )
                )
            except Exception as exc:
                missing = set(
                    key
                    for key in allkeys
                    if key not in list(self.config.__dict__.keys())
                ) - set(config.keys())
                exc.args = errors.addmessage(
                    exc.args,
                    ". Unable to parse configuration option '{0}'. The error can quite likely be solved by modifying the option in the configuration file.".format(
                        ",".join(missing)
                    ),
                )
                raise
            for k in config:
                print(
                    "warning: unrecognized configuration option {0} for {1}".format(
                        k, self.__class__.__name__
                    )
                )
            self.config.class_ = self.__class__

    def parse_config(self, config):
        # every known option should be pop()'ed from config, converted to a
        # proper type and stored as property in self.config, for example:
        # self.config.foo = int(config.pop('foo', 1))
        pass


### FILES
def best_effort_atomic_rename(src, dest):
    if sys.platform == "win32" and os.path.exists(dest):
        os.remove(dest)
    os.rename(src, dest)


def filename_enumerator(filename, start=0):
    base, ext = os.path.splitext(filename)
    for count in itertools.count(start):
        yield "{0}_{2}{1}".format(base, ext, count)


def find_unused_filename(filename):
    if not os.path.exists(filename):
        return filename
    for f in filename_enumerator(filename, 2):
        if not os.path.exists(f):
            return f


def label_enumerator(label, start=0):
    for count in itertools.count(start):
        yield "{0}_{1}".format(label, count)


def find_unused_label(label, labellist):
    for l in label_enumerator(label):
        if not l in labellist:
            return l


def yield_when_exists(filelist, timeout=None):
    """Wait for files in 'filelist' to appear, for a maximum of 'timeout' seconds,
    yielding them in arbitrary order as soon as they appear.
    If 'filelist' is a set, it will be modified in place, and on timeout it will
    contain the files that have not appeared yet."""
    if not isinstance(filelist, set):
        filelist = set(filelist)
    delay = loop_delayer(5)
    start = time.time()
    while filelist:
        next(delay)
        exists = set(f for f in filelist if os.path.exists(f))
        for e in exists:
            yield e
        filelist -= exists
        if timeout is not None and time.time() - start > timeout:
            break


def wait_for_files(filelist, timeout=None):
    """Wait until the files in 'filelist' have appeared, for a maximum of 'timeout' seconds.
    Returns True on success, False on timeout."""
    filelist = set(filelist)
    for i in yield_when_exists(filelist, timeout):
        pass
    return not filelist


def wait_for_file(file, timeout=None):
    return wait_for_files([file], timeout=timeout)


def space_to_edf(space, filename):
    header = {}
    for a in space.axes:
        header[str(a.label)] = "{0} {1} {2}".format(a.min, a.max, a.res)
    edf = EdfFile.EdfFile(filename)
    edf.WriteImage(header, space.get_masked().filled(0), DataType="Float")


def space_to_txt(space, filename):
    data = [coord.flatten() for coord in space.get_grid()]
    data.append(space.get_masked().filled(0).flatten())
    data = numpy.array(data).T

    with open(filename, "w") as fp:
        fp.write("\t".join(ax.label for ax in space.axes))
        fp.write("\tintensity\n")
        numpy.savetxt(fp, data, fmt="%.6g", delimiter="\t")


@contextlib.contextmanager
def open_h5py(fn, mode):
    if isinstance(fn, h5py._hl.group.Group):
        yield fn
    else:
        with h5py.File(fn, mode) as fp:
            if mode == "w":
                fp.create_group("binoculars")
                yield fp["binoculars"]
            if mode == "r":
                if "binoculars" in fp:
                    yield fp["binoculars"]
                else:
                    yield fp


### VARIOUS


def uniqid():
    return "{0:08x}".format(random.randint(0, 2 ** 32 - 1))


def grouper(iterable, n):
    while True:
        chunk = list(itertools.islice(iterable, n))
        if not chunk:
            break
        yield chunk


_python_executable = None


def register_python_executable(scriptname):
    global _python_executable
    _python_executable = sys.executable, scriptname


def get_python_executable():
    return _python_executable


def chunk_slicer(count, chunksize):
    """yields slice() objects that split an array of length 'count' into equal sized chunks of at most 'chunksize'"""
    chunkcount = int(numpy.ceil(float(count) / chunksize))
    realchunksize = int(numpy.ceil(float(count) / chunkcount))
    for i in range(chunkcount):
        yield slice(i * realchunksize, min(count, (i + 1) * realchunksize))


def cluster_jobs(jobs, target_weight):
    jobs = sorted(jobs, key=lambda job: job.weight)

    # we cannot split jobs here, so just yield away all jobs that are overweight or just right
    while jobs and jobs[-1].weight >= target_weight:
        yield [jobs.pop()]

    while jobs:
        cluster = [jobs.pop()]  # take the biggest remaining job
        size = cluster[0].weight
        for i in range(
            len(jobs) - 1, -1, -1
        ):  # and exhaustively search for all jobs that can accompany it (biggest first)
            if size + jobs[i].weight <= target_weight:
                size += jobs[i].weight
                cluster.append(jobs.pop(i))
        yield cluster


def cluster_jobs2(jobs, target_weight):
    """Taking the first n jobs that together add up to target_weight.
       Here as opposed to cluster_jobs the total number of jobs does not have to be known beforehand
    """
    jobslist = []
    for job in jobs:
        jobslist.append(job)
        if sum(j.weight for j in jobslist) >= target_weight:
            yield jobslist[:]
            jobslist = []
    if len(jobslist) > 0:  # yield the remainder of the jobs
        yield jobslist[:]


def loop_delayer(delay):
    """Delay a loop such that it runs at most once every 'delay' seconds. Usage example:
    delay = loop_delayer(5)
    while some_condition:
        next(delay)
        do_other_tasks
    """

    def generator():
        polltime = 0
        while 1:
            diff = time.time() - polltime
            if diff < delay:
                time.sleep(delay - diff)
            polltime = time.time()
            yield

    return generator()


def transformation_from_expressions(space, exprs):
    def transformation(*coords):
        ns = dict((i, getattr(numpy, i)) for i in dir(numpy))
        ns.update(**dict((ax.label, coord) for ax, coord in zip(space.axes, coords)))
        return tuple(eval(expr, ns) for expr in exprs)

    return transformation


def format_bytes(bytes):
    units = "kB", "MB", "GB", "TB"
    exp = min(max(int(numpy.log(bytes) / numpy.log(1024.0)), 1), 4)
    return "{0:.1f} {1}".format(bytes / 1024 ** exp, units[exp - 1])


### GZIP PICKLING (zpi)

# handle old zpi's
def _pickle_translate(module, name):
    if module in ("__main__", "ivoxoar.space") and name in ("Space", "Axis"):
        return "BINoculars.space", name
    return module, name


if inspect.isbuiltin(pickle.Unpickler):
    # real cPickle: cannot subclass
    def _find_global(module, name):
        module, name = _pickle_translate(module, name)
        __import__(module)
        return getattr(sys.modules[module], name)

    def pickle_load(fileobj):
        unpickler = pickle.Unpickler(fileobj)
        unpickler.find_global = _find_global
        return unpickler.load()


else:
    # pure python implementation
    class _Unpickler(pickle.Unpickler):
        def find_class(self, module, name):
            module, name = _pickle_translate(module, name)
            return pickle.Unpickler.find_class(self, module, name)

    def pickle_load(fileobj):
        unpickler = _Unpickler(fileobj)
        return unpickler.load()


@contextlib.contextmanager
def atomic_write(filename):
    """Atomically write data into 'filename' using a temporary file and os.rename()

    Rename on success, clean up on failure (any exception).

    Example:
    with atomic_write(filename) as tmpfile
        with open(tmpfile, 'w') as fp:
            fp.write(...)
    """

    if isinstance(filename, h5py._hl.group.Group):
        yield filename
    else:
        tmpfile = "{0}-{1}.tmp".format(os.path.splitext(filename)[0], uniqid())
        try:
            yield tmpfile
        except:
            raise
        else:
            best_effort_atomic_rename(tmpfile, filename)
        finally:
            if os.path.exists(tmpfile):
                os.remove(tmpfile)


def zpi_save(obj, filename):
    with atomic_write(filename) as tmpfile:
        fp = gzip.open(tmpfile, "wb")
        try:
            pickle.dump(obj, fp, pickle.HIGHEST_PROTOCOL)
        finally:
            fp.close()


def zpi_load(filename):
    if hasattr(filename, "read"):
        fp = gzip.GzipFile(filename.name, fileobj=filename)
    else:
        fp = gzip.open(filename, "rb")
    try:
        return pickle_load(fp)
    finally:
        fp.close()


def serialize(space, command):
    # first 48 bytes contain length of the message, whereby the first 8 give the length of the command, the second 8 the length of the configfile etc..
    message = io.StringIO()
    message.write(struct.pack("QQQQQQ", 0, 0, 0, 0, 0, 0))

    message.write(command)
    commandlength = message.len - 48

    message.write(space.config.serialize())
    configlength = message.len - commandlength - 48

    message.write(space.metadata.serialize())
    metalength = message.len - configlength - commandlength - 48

    numpy.save(message, space.axes.toarray())
    arraylength = message.len - metalength - configlength - commandlength - 48

    numpy.save(message, space.photons)
    photonlength = (
        message.len - arraylength - metalength - configlength - commandlength - 48
    )

    numpy.save(message, space.contributions)
    contributionlength = (
        message.len
        - photonlength
        - arraylength
        - metalength
        - configlength
        - commandlength
        - 48
    )

    message.seek(0)
    message.write(
        struct.pack(
            "QQQQQQ",
            commandlength,
            configlength,
            metalength,
            arraylength,
            photonlength,
            contributionlength,
        )
    )
    message.seek(0)

    return message


def packet_slicer(length, size=1024):  # limit the communication to 1024 bytes
    while length > size:
        length -= size
        yield size
    yield length


def socket_send(ip, port, mssg):
    try:
        mssglengths = struct.unpack(
            "QQQQQQ", mssg.read(48)
        )  # the lengths of all the components
        mssg.seek(0)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((ip, port))

        sock.send(mssg.read(48))
        for l in mssglengths:
            for packet in packet_slicer(l):
                sock.send(mssg.read(packet))
        sock.close()
    except socket.error:  # in case of failure to send. The data will be saved anyway so any loss of communication unfortunate but not critical
        pass


def socket_recieve(RequestHandler):  # pass one the handler to deal with incoming data
    def get_msg(length):
        msg = io.StringIO()
        for packet in packet_slicer(length):
            p = RequestHandler.request.recv(
                packet, socket.MSG_WAITALL
            )  # wait for full mssg
            msg.write(p)
        if msg.len != length:
            raise errors.CommunicationError(
                "recieved message is too short. expected length {0}, recieved length {1}".format(
                    length, msg.len
                )
            )
        msg.seek(0)
        return msg

    command, config, metadata, axes, photons, contributions = tuple(
        get_msg(msglength)
        for msglength in struct.unpack(
            "QQQQQQ", RequestHandler.request.recv(48, socket.MSG_WAITALL)
        )
    )
    return (
        command.read(),
        config.read(),
        metadata.read(),
        numpy.load(axes),
        numpy.load(photons),
        numpy.load(contributions),
    )
